package ca.mcgill.lmp.finite;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Bisim encapsulates all the information required to run the bisim() algorithm. It
 * also contains the method getLumpedLMP() that build a lumped LMP using the 
 * information produced by bisim(). A common usage would be: <br />
 * 
 * <code>
 * Bisim bisim = new Bisim(myLMP); <br />
 * bisim.bisim(); <br />
 * LMP lumpedLMP = bisim.getLumpedLMP; <br />
 * System.out.println("result: " + lumpedLMP.toString(); <br />
 * </code>
 * 
 * @author Alexandre Bouchard
 */
public class Bisim 
{
	/**
	 * 
	 */
    private LMP lmp;
	
	/**
	 * Global variable needed for bisim().
	 * Indexed by the sets of D1, map them to the formula (a String) that 
	 * all the states in the given set satisfy.
	 */
	private Map formulas = new HashMap();
	
	/**
	 * Global variables needed for bisim().
	 * Contains sets of states
	 */
	private Set D1, D2;
	
	/**
	 * Create an instance of LMP to lump the given lmp. This lmp will not be 
	 * modified.
	 * @param lmp
	 */
	public Bisim(LMP lmp)
	{
		this.lmp = lmp;
	}
	
	/**
	 * This algorithm allows us to distinguish states that do not satisfy the same 
	 * formulas in the following logic: <br />
	 * <code>
	 * L0 :== T | f_1 ^ f_2 | <a>q f <br />
	 * </code>
	 * This logic has the property to characterize bisimulation: two states are 
	 * bisimilar iff they satisfy the same formulas in L0.
	 * Beginning with D1 containing only the set of all states, bisim operates as 
	 * follow: for each label a in the set of labels and B2 in D1, it ``splits'' 
	 * every set B1 of D1 into subsets having the same probability of jumping to B2 
	 * with action a; this is done until it does not modify D1. At the end, all 
	 * states satisfying the same formulas will belong to exactly the same sets in 
	 * D1.
	 * <br /><br />
	 * From: J. Desharnais, A. Edalat, P. Panangaden, Bisimulation for Labelled 
	 * Markov Processes
	 *
	 */
	public void bisim()
	{
		D1 = new HashSet();
		D1.add(lmp.getStates());
		formulas.put(lmp.getStates(), "T");
		D2 = new HashSet();
		while (!D1.equals(D2))
		{
			Iterator labelsIterator = lmp.getLabels().iterator();	
			while (labelsIterator.hasNext())
			{
				Label currentLabel = (Label) labelsIterator.next();
				Iterator stateSubsetsIterator2 = D1.iterator();
				while (stateSubsetsIterator2.hasNext())
				{
					Set B2 = (Set) stateSubsetsIterator2.next();
					D2 = D1;
					D1 = new HashSet();
					Iterator stateSubsetsIterator1 = D2.iterator();
					while (stateSubsetsIterator1.hasNext())
					{
						Set B1 = (Set) stateSubsetsIterator1.next();					
						D1.addAll(split(B1, currentLabel, B2));
					}
				}
			}
		}
	}
	
	/**
	 * See bisim().
	 * @return The set D1 that bisim() computed.
	 */
	public Set getD1()
	{
	    if (D1 == null)
	    {
	        throw new RuntimeException("bisim() must be invoqued before calling getD1.");
	    }
	    else
	    {
	        return D1;
	    }
	}
	
	/**
	 * A way to obtain the formulas as a map indexed by the sets of D1 and mapped to 
	 * the formula in L0 (a String) that all the states in the given set satisfy. <br />
	 * See bisim() and split().
	 * @return The map of formulas
	 */
	public Map getFormulas()
	{
	    if (formulas == null)
	    {
	        throw new RuntimeException("bisim() must be invoqued before calling getFormulas.");
	    }
	    else
	    {
	        return formulas;
	    }
	}
	
	/**
	 * split() first sorts all the states in B1 by P_a(b, B2) for b in B1. Then, for 
	 * every nonzero value of P_a(b, B2) in this order, the states in B1 that can 
	 * jump into B2 with probability at least this value are added to the returned 
	 * set which is initialized to contains only B1 at the beginning. Before adding 
	 * a set to D, split() checks if it is already a member of D, D1, or D2; if not, 
	 * it defines the formula represented by the set and adds the set to D; 
	 * otherwise, it does not add the set to D. This ensure that we don't compute 
	 * a new formula for a set that already has a shorter one.
	 * <br /><br />
	 * From: J. Desharnais, A. Edalat, P. Panangaden, Bisimulation for Labelled 
	 * Markov Processes
	 * 
	 * @param B1 A set of states.
	 * @param label A label.
	 * @param B2 Another set of states.
	 * @return The set D as computed in the above description.
	 */
	private Set split(Set B1, Label label, Set B2)
	{
		List orderedStates = new ArrayList(B1);
		Set D = new HashSet();
		D.add(B1);
		Collections.sort(orderedStates, new ProbabilityComparator(label, B2));
		double thresholdProbability = 0;
		for (int i = 0; i < orderedStates.size(); i++)
		{
			double currentProbability = computeProbability(label, (State) orderedStates.get(i), B2);
			if (currentProbability > thresholdProbability)
			{
				thresholdProbability = currentProbability;
				Set B3 = new HashSet(orderedStates.subList(i, orderedStates.size()));
				if ( !D.contains(B3) && !D1.contains(B3) && !D2.contains(B3) )
				{
					D.add(B3);
					formulas.put(B3,	(String) formulas.get(B1) + 
										" ^ <" + label.toString() + ">(" + currentProbability + ")" +
										"{" + (String) formulas.get(B2) + "}");
				}
			}
		}
		return D;
	}
	
	private double computeProbability(Label label, State state, Collection target)
	{
		return lmp.getTPM(label).Pr(state, target);
	}
	
	/**
	 * ProbabilityComparator is used by the split() method to sort states using 
	 * Collections.sort(list, comparator). It compares 2 State's s_1 and s_2 by 
	 * comparing the values P_label(s_1, target) and P_label(s_2, target)
	 *
	 */
	class ProbabilityComparator implements Comparator
	{
		private Label label;
		private Set target;
		
		public ProbabilityComparator(Label label, Set target)
		{
			this.label = label;
			this.target = target;
		}
	
		public int compare(Object object1, Object object2) 
		{
			State state1 = (State) object1;
			State state2 = (State) object2;
			double pr1 = computeProbability(label, state1, target);
			double pr2 = computeProbability(label, state2, target);
			if (pr1 < pr2)
			{
				return -1;
			}
			else if (pr1 > pr2)
			{
				return 1;
			}
			else
			{
				return 0;
			}
		}
		
	}
	
	/**
	 * After the execution of bisim(), Bisim has all the infomation required to compute
	 * a simplified, "lumped" LMP. 
	 * Indeed, the sets in D1=D2 represent (after the execution of bisim) formulas in a 
	 * logic s.t. 2 states are bisimilar iff they belong to the sets in D1. 
	 * So the sets in D1 can be viewed as binary partitions of the states, and the first 
	 * step to compute the lumped LMP is to compute the union of those partitions.
	 * The second step is to compute the transition matrices with straightforward additions 
	 * of probabilities.
	 * A new LMP is finally created with those informations and returned.
	 * <br />
	 * Note: A rather suboptimal way to do that but we don't care about the time 
	 * complexity and implementation of this method. "Premature optimization is the 
	 * root of all evil." --D. Knuth
	 * 
	 * @return The lumped LMP
	 */
	public LMP getLumpedLMP()
	{
		// Compute the union of the partitions induced by sets in D1
		Set lumpedStates1 = new HashSet();
		lumpedStates1.add(lmp.getStates());
		Iterator partitionsIterator = D1.iterator();
		while (partitionsIterator.hasNext())
		{
			Set currentPartition = (Set) partitionsIterator.next();
			Set lumpedStates2 = new HashSet();
			lumpedStates2.addAll(lumpedStates1);
			Iterator lumpedStatesIterator = lumpedStates2.iterator();
			while (lumpedStatesIterator.hasNext())
			{
				Set currentLumpedState = (Set) lumpedStatesIterator.next();
				partitionUnion(lumpedStates1, currentLumpedState, currentPartition);
			}
		}
		
		// Build the set of states for the lumped LMP.
		Iterator statesIterator = lumpedStates1.iterator();
		Set states = new HashSet();
		List listOfLumpedStates = new ArrayList(lumpedStates1.size());
		int index = 0;
		while (statesIterator.hasNext())
		{
			Set currentLumpedState = (Set) statesIterator.next();
			State currentState = new State(index, currentLumpedState.toString());
			states.add(currentState);
			listOfLumpedStates.add(currentLumpedState);
			index++;
		}
		
		// Compute the new transition matrices
		List transitionMatrices = new ArrayList(lmp.getNumberOfLabels());
		Iterator labelIterator = lmp.getLabels().iterator();
		while (labelIterator.hasNext())
		{
			Label currentLabel = (Label) labelIterator.next();
			double [][] transitionProbabilities = new double[lumpedStates1.size()][];
			for (int i = 0; i < lumpedStates1.size(); i++)
			{
				transitionProbabilities[i] = new double[lumpedStates1.size()];
			}
			Iterator i = states.iterator();
			while (i.hasNext())
			{
				State currentState1 = (State) i.next();
				Iterator j = states.iterator();
				while (j.hasNext())
				{
					State currentState2 = (State) j.next();
					State currentState3 = (State) (((Set) listOfLumpedStates.get(currentState1.getIndex())).iterator().next());
					transitionProbabilities[currentState1.getIndex()][currentState2.getIndex()] = 
					    computeProbability(currentLabel, currentState3, (Set) listOfLumpedStates.get(currentState2.getIndex()));
				}
			}
			transitionMatrices.add(new TransitionMatrix(transitionProbabilities));
		}
		
		// The lumped LMP is created and returned
		return new LMP(states, lmp.getLabels(), transitionMatrices);
	}	
	
	/**
	 * The set lumpedState in the set of sets lumpedStates is split according to the 
	 * binary partition induce by the set partition. <br />
	 * See: getLumpedLMP().
	 * @param lumpedStates
	 * @param lumpedState
	 * @param partition
	 */
	private void partitionUnion(Set lumpedStates, Set lumpedState, Set partition)
	{
		lumpedStates.remove(lumpedState);
		Set insidePartition = new HashSet();
		Set outsidePartition = new HashSet();
		Iterator stateIterator = lumpedState.iterator();
		while (stateIterator.hasNext())
		{
			Object currentState = stateIterator.next();
			if (partition.contains(currentState))
			{
				insidePartition.add(currentState);
			}
			else
			{
				outsidePartition.add(currentState);
			}
		}
		if (!insidePartition.isEmpty())
		{
			lumpedStates.add(insidePartition);
		}
		if (!outsidePartition.isEmpty())
		{
			lumpedStates.add(outsidePartition);
		}
	}
	
	/**
	 * <code>bisim [optionalArguments] &lt;inputFile&gt;</code>
	 * <br />
	 * Bisim takes the description of a LMP as input and outputs
	 * a LMP obtained by quotienting the input LMP by the
	 * equivalence relation of bisimulation
	 * <br />
	 * First argument: the path to a file describing a LMP <br />
	 * Optional arguments: <br />
	 * &nbsp;&nbsp;--transitions  Output the transition matrices of the new LMP <br />
	 * &nbsp;&nbsp;--sets  Output the set D1 used in an intermediate step <br />
	 * The LMP should be described using the following specification: <br />
	 * The description should be a list of transition matrices
	 * (one per label) separated by the "\n*\n" string. The
	 * transition matrices should be formated by separating columns
	 * by tabs and rows by new lines characters.
	 * @param args
	 */
	public static void main(String [] args)
	{  
	    if (args.length == 0 || 
	            (args.length == 1 && args[0].equals("-h")) ||
	            (args.length == 1 && args[0].equals("--help")) ||
	            (args.length == 1 && args[0].equals("--usage")))
	    {
	        printUsage();
	        return;
	    }	    
	    
	    boolean transitions = false;
	    boolean sets = false;
	    Set optionalArguments = new HashSet();
	    for (int i = 0; i < args.length - 1; i++)
	    {
	        optionalArguments.add(args[i]);
	    }
	    if (optionalArguments.contains("--transitions"))
	    {
	        transitions = true;
	    }
	    if (optionalArguments.contains("--sets"))
	    {
	        sets = true;
	    }
	    
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(args[args.length - 1]));
			StringBuffer buffer = new StringBuffer();
			while (in.ready())
			{
				buffer.append(in.readLine() + "\n");
			}
			String kernelDescription = buffer.toString();
			LMP lmp = new LMP(kernelDescription);
			System.out.println("Input:");
			System.out.println(lmp.toString());
			Bisim bisim = new Bisim(lmp);
			bisim.bisim();
			LMP lumped = bisim.getLumpedLMP();
			System.out.println("Output (lumped LMP):");
			System.out.println(lumped.toString());
			if (transitions)
			{
			    System.out.println("Transition matrices of the lumped LMP:");
			    Iterator labelsIterator = lumped.getLabels().iterator();
			    while (labelsIterator.hasNext())
			    {
			        System.out.println(lumped.getTPM((Label) labelsIterator.next()) + "\n");
			    }
			}
			if (sets)
			{
			    System.out.println("Sets in D1:");
			    System.out.println(bisim.D1.toString());
			}
		}
		catch (FileNotFoundException fne)
		{
			System.err.println("Input file not found: double check the path.");
		}
		catch (IOException ioe)
		{
			System.err.println("An error occured while trying to read the input file.");
		}
		catch (Exception e)
		{
		    System.err.println("An internal error occured.");
		    System.err.println("Description: " + e.toString());
		}
	}
	
	/**
	 * Print in System.err a description of the functionalities of main().
	 * <br />
	 * See main().
	 *
	 */
	public static void printUsage()
	{
	    System.err.println("bisim [optionalArguments] <inputFile>");
	    System.err.println("Bisim takes the description of a LMP as input and outputs");
	    System.err.println("a LMP obtained by quotienting the input LMP by the ");
	    System.err.println("equivalence relation of bisimulation.");
	    System.err.println("First argument: the path to a file describing a LMP");
	    System.err.println("Optional arguments:");
	    System.err.println("\t--transitions  Output the transition matrices of the new LMP");
	    System.err.println("\t--sets  Output the set D1 used in an intermediate step (for");
	    System.err.println("\t               debug, see sources of details)");
	    System.err.println("The LMP should be described using the following specification:");
	    System.err.println("The description should be a list of transition matrices");
	    System.err.println("(one per label) separated by the \"\\n*\\n\" string. The");
	    System.err.println("transition matrices should be formated by separating columns");
	    System.err.println("by tabs and rows by new lines characters.");
	}
}
