package ca.mcgill.lmp.finite;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * A representation of a finite LMP.
 * 
 * @author Alexandre Bouchard
 */
public class LMP 
{
	private Set states, labels;
	
	private List transitionMatrices;
	
	/**
	 * Construct a new LMP using "\n*\n"-separated description of transition matrices
	 * (See TransitionMatrix.java). e.g. <br />
	 * 0&nbsp;&nbsp;1&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;1<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * *<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;0<br />
	 * 0&nbsp;&nbsp;0&nbsp;&nbsp;0&nbsp;&nbsp;1<br />
	 * will produced a non-lumpable LMP.
	 * @param description
	 */
	public LMP(String description)
	{
		String [] transitionDescriptions = description.split("\\*\\n");
		transitionMatrices = new ArrayList();
		for (int i = 0; i < transitionDescriptions.length; i++)
		{
			transitionMatrices.add(new TransitionMatrix(transitionDescriptions[i]));
		}
		initStates();
		initLabels();
	}
	
	/**
	 * 
	 * @param states
	 * @param labels
	 * @param transitionMatrices
	 */
	public LMP(Set states, Set labels, List transitionMatrices)
	{
		this.transitionMatrices = transitionMatrices;
		this.states = states;
		this.labels = labels;
	}
	
	/**
	 * 
	 *
	 */
	private void initStates()
	{
		states = new HashSet();
		for (int i = 0; i < getNumberOfStates(); i++)
		{
			states.add(new State(i));
		}		
	}
	
	/**
	 * 
	 *
	 */
	private void initLabels()
	{
		labels = new HashSet();
		for (int i = 0; i < getNumberOfLabels(); i++)
		{
			labels.add(new Label(i));
		}		
	}
	
	/**
	 * 
	 * @return
	 */
	public int getNumberOfStates()
	{
		return ((TransitionMatrix) transitionMatrices.get(0)).size();
	}
	
	/**
	 * 
	 * @return
	 */
	public int getNumberOfLabels()
	{
		return transitionMatrices.size();
	}

	/**
	 * @return
	 */
	public Set getLabels() 
	{
		return labels;
	}

	/**
	 * @return
	 */
	public Set getStates() 
	{
		return states;
	}
	
	/**
	 * 
	 * @param label
	 * @return the transition probability matrix corresponding to the provided label.
	 */
	public TransitionMatrix getTPM(Label label)
	{
		return (TransitionMatrix) transitionMatrices.get(label.getIndex());
	}
	
	/**
	 * 
	 * @return
	 */
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("Number of states: " + getNumberOfStates() + "\n");
		buffer.append("States: ");
		Iterator iterator = states.iterator();
		while (iterator.hasNext())
		{
			buffer.append(iterator.next().toString() + "  ");
		}
		
		return buffer.toString();
	}
}
