package ca.mcgill.lmp.finite;

import java.util.Collection;
import java.util.Iterator;

/**
 * A transition matrix.
 * 
 * @author Alexandre Bouchard
 */
public class TransitionMatrix 
{
	/**
	 * The matrix of probabilities.
	 */
    private double [][] transitionProbabilities;
	
	/**
	 * The number of states of the underlying probability space.
	 * @return
	 */
    public int size()
	{
		return transitionProbabilities.length;
	}
	
	/**
	 * Create a new transition probabilities based on a String description.
	 * The String should contain a matrix with columns separated by tabs and rows
	 * separated by new lines. Entry i,j should represent the probability to go from 
	 * state i to state j.
	 * @param description
	 */
    public TransitionMatrix(String description)
	{
		String [] lines = description.split("\\n");
		transitionProbabilities = new double[lines.length][];
		for (int i = 0; i < lines.length; i++)
		{
			transitionProbabilities[i] = new double[lines.length];
		}
		for (int i = 0; i < lines.length; i++)
		{
			String [] entries = lines[i].split("\\t");
			if (entries.length != lines.length)
			{
			    throw new RuntimeException("The transition matrix should be a square matrix.");
			}
			for (int j = 0; j < entries.length; j++)
			{
				transitionProbabilities[i][j] = Double.parseDouble(entries[j]);
			}
		}
	}
	
	/**
	 * Create a new transition probability matrix with the provided probabilities.
	 * transitionProbilities[i][j] should be the probability to go from state i to 
	 * state j.
	 * @param transitionProbabilities
	 */
    public TransitionMatrix(double [][] transitionProbabilities)
	{
		this.transitionProbabilities = transitionProbabilities;
	}
		
	/**
	 * 
	 * @param from The initial state.
	 * @param to The destination state.
	 * @return The probability to go from the initial to the destination state.
	 */
    public double Pr(State from, State to)
	{
		return transitionProbabilities[from.getIndex()][to.getIndex()];
	}
		
	/**
	 * 
	 * @param from The initial state.
	 * @param to A set of States, the destination states.
	 * @return The probability to transition from the initial state to any one of 
	 * the state in the provided set of destination states.
	 */
    public double Pr(State from, Collection to)
	{
		Iterator i = to.iterator();
		double partialSum = 0;
		while (i.hasNext())
		{
			State currentState = (State) i.next();
			partialSum = partialSum + Pr(from, currentState);
		}
		return partialSum;
	}
	
	/**
	 * 
	 * @return
	 */
    public String toString()
	{
		StringBuffer buffer = new StringBuffer();
		for (int i = 0; i < size(); i++)
		{
			buffer.append("|\t");
			for (int j = 0; j < size(); j++)
			{
				buffer.append("" + transitionProbabilities[i][j] + "\t\t");
			}
			buffer.append("|\n");
		}
		return buffer.toString();
	}
}
